//##############################################################################################################//
#include "CmnHdr.h"

#include <windows.h>
#include <windowsX.h>
#include <process.h>
#include <tchar.h>
#include <Winsock2.h>

#include "resource.h"

//###############################################################################################################//
#pragma comment( lib, "ws2_32" )

//###############################################################################################################//
#define MAX_THREADS 1000 // Threads array size.

//###############################################################################################################//
HWND   g_hwndMain              = NULL;
CHAR   g_szServerIP[32]        = {0};
DWORD  g_dwServerPort          = 0;
LONG   g_lConnect              = 0;
LONG   g_lSend                 = 0;
LONG   g_lRecv                 = 0;
LONG   g_lError                = 0;
DWORD  g_dwThreadsCount        = 0;
HANDLE g_hThreads[MAX_THREADS] = {0};
volatile LONG g_fContinue = 1;

//###############################################################################################################//
VOID LogError( PSTR pszFunc, DWORD dwError )
{
    HANDLE hFile = CreateFile( TEXT("EchoClient.err"), GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, 
						       OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
    if( hFile == INVALID_HANDLE_VALUE ){ return; }

    SetFilePointer( hFile, 0, NULL, FILE_END );

    CHAR szError[32] = {0};
    wsprintfA( szError, "%s: %d\r\n", pszFunc, dwError );

    DWORD dwWrite = 0;
    WriteFile( hFile, szError, lstrlenA( szError ), &dwWrite, NULL );

    CloseHandle( hFile );

    InterlockedIncrement( &g_lError );
    PostMessage( g_hwndMain, WM_USER + 1, NULL, NULL );
}

//###############################################################################################################//
DWORD WINAPI Thread( PVOID pvSome )
{
    BYTE         bBuffer[BUF_SIZE] = {0};
    SOCKET       Socket            = INVALID_SOCKET;
    sockaddr_in  server            = {0};
    hostent     *host              = NULL;
    int          iRet              = 0;

    server.sin_family      = AF_INET;
    server.sin_port        = htons( g_dwServerPort );
    server.sin_addr.s_addr = inet_addr( g_szServerIP );

    while( g_fContinue ){
        __try{
            Socket = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
            if( INVALID_SOCKET == Socket ){
                LogError( "socket()", WSAGetLastError() );
                __leave;
                }

            if( connect( Socket, (sockaddr*) &server, sizeof( server ) ) == SOCKET_ERROR ){
                LogError( "connect()", WSAGetLastError() );
                __leave;
                }
            InterlockedIncrement( &g_lConnect );
            PostMessage( g_hwndMain, WM_USER + 1, NULL, NULL );

            // Send 1Kb.
            iRet = send( Socket, (char*) bBuffer, 1024, 0 );
            if( SOCKET_ERROR == iRet ){
                LogError( "send()", WSAGetLastError() );
                __leave;
                }
            InterlockedIncrement( &g_lSend );
            PostMessage( g_hwndMain, WM_USER + 1, NULL, NULL );

            iRet = recv( Socket, (char*) bBuffer, BUF_SIZE, 0 );
            if( SOCKET_ERROR == iRet ){
                LogError( "recv()", WSAGetLastError() );
                __leave;
                }
            InterlockedIncrement( &g_lRecv );
            PostMessage( g_hwndMain, WM_USER + 1, NULL, NULL );
            }
        __finally{
            if( Socket != INVALID_SOCKET ){
                shutdown( Socket, SD_BOTH );
                closesocket( Socket );
                Socket = INVALID_SOCKET;
                }
            }
        }

    return 0;
}

//###############################################################################################################//
void DlgMain_OnCommand( HWND hWnd, int id, HWND hWndCtl, UINT codeNotify )
{
	switch( id ){
        case ID_BUTTON_START:{
            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_START ), FALSE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_SERVER  ), FALSE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_PORT    ), FALSE );

            GetDlgItemTextA( hWnd, ID_EDIT_SERVER, g_szServerIP, sizeof( g_szServerIP ) / sizeof( g_szServerIP[0] ) );
            g_dwServerPort = GetDlgItemInt( hWnd, ID_EDIT_PORT, NULL, FALSE );

            g_lConnect = 0;
            g_lSend    = 0;
            g_lRecv    = 0;
            g_lError   = 0;

            SYSTEM_INFO sysinfo = {0};
            GetSystemInfo( &sysinfo );

            g_dwThreadsCount = sysinfo.dwNumberOfProcessors * 2;

            if( g_dwThreadsCount > MAX_THREADS ){
                MessageBox( NULL, TEXT("Too many processors ;)"), TEXT("EchoClient"), MB_OK );
                EnableWindow( GetDlgItem( hWnd, ID_BUTTON_START ), TRUE );
                EnableWindow( GetDlgItem( hWnd, ID_EDIT_SERVER  ), TRUE );
                EnableWindow( GetDlgItem( hWnd, ID_EDIT_PORT    ), TRUE );
                break;
                }

            // Start threads.
            for( int i = 0; i < g_dwThreadsCount; i++ ){
                g_hThreads[i] = BEGINTHREAD( Thread, NULL );
                }

            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_STOP ), TRUE );
			break;
            }

        case ID_BUTTON_STOP:
            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_STOP ), FALSE );

            InterlockedExchange( &g_fContinue, 0 );

            WaitForMultipleObjects( g_dwThreadsCount, g_hThreads, TRUE, INFINITE );

            for( int i = 0; i < g_dwThreadsCount; i++ ){
                CloseHandle( g_hThreads[i] );
                }

            g_fContinue = 1;

            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_START ), TRUE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_SERVER  ), TRUE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_PORT    ), TRUE );
			break;

        case ID_CHECK_ONTOP:
            SetWindowPos( hWnd, IsDlgButtonChecked( hWnd, ID_CHECK_ONTOP ) ? HWND_TOPMOST : HWND_NOTOPMOST,
                          0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE );
            break;
        }
}

//###############################################################################################################//
BOOL DlgMain_OnInitDialog( HWND hWnd, HWND hWndFocus, LPARAM lParam )
{
    g_hwndMain = hWnd;

    SetDlgItemText( hWnd, ID_EDIT_SERVER, TEXT("127.0.0.1") );
    SetDlgItemInt( hWnd, ID_EDIT_PORT, 6060, FALSE );

    WSADATA wsd = {0};
    WSAStartup( MAKEWORD( 2,2 ), &wsd );

    return TRUE;
}

//###############################################################################################################//
void DlgMain_OnClose( HWND hWnd )
{
    WSACleanup();
	EndDialog( hWnd, 0 );
}

//###############################################################################################################//
int WINAPI Dialog_Main( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
    switch( message ){
        Dlg_Proc( hWnd, WM_INITDIALOG, DlgMain_OnInitDialog );
        Dlg_Proc( hWnd, WM_COMMAND,    DlgMain_OnCommand    );
        Dlg_Proc( hWnd, WM_CLOSE,      DlgMain_OnClose      );

        case WM_USER + 1:
            SetDlgItemInt( hWnd, ID_STATIC_CONNECT, g_lConnect, FALSE );
            SetDlgItemInt( hWnd, ID_STATIC_SEND,    g_lSend,    FALSE );
            SetDlgItemInt( hWnd, ID_STATIC_RECV,    g_lRecv,    FALSE );
            SetDlgItemInt( hWnd, ID_STATIC_ERROR,   g_lError,   FALSE );
            break;
        }

    return 0;
}

//##############################################################################################################//
int WINAPI _tWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow )
{
    DialogBox( hInstance, MAKEINTRESOURCE( IDD_DIALOG_MAIN ), NULL, Dialog_Main );

    return 0;
}

//##############################################################################################################//