//##############################################################################################################//
/*
    Proxy-EchoServer example.

    This example shows how to connect to other servers in the asynchronous
    mode and exchange data with them in the same way as with clients.

 */
//##############################################################################################################//
#include "CmnHdr.h"

#include <windows.h>
#include <windowsX.h>
#include <tchar.h>

#include "CPServer.h"

#include "resource.h"

//##############################################################################################################//
CPServer PServer; // ProstoServer class.

HWND  g_hwndMain                         = NULL;
BOOL  g_fStarted                         = FALSE;
CHAR  g_szServerIP[32]                   = {0};
DWORD g_dwServerPort                     = 0;
LONG  g_lOnConnectCount                  = 0;
LONG  g_lOnOutgoingConnectCount          = 0;
LONG  g_lOnDisconnectCount               = 0;
LONG  g_lOnOutgoingDisconnectCount       = 0;
LONG  g_lOnRemoveConnectionCount         = 0;
LONG  g_lOnOutgoingRemoveConnectionCount = 0;

// Common structure for an incoming and outgoing connection.
typedef struct _COMMON_DATA{
    BYTE        bIncomingBuffer[BUF_SIZE];
    BYTE        bOutgoingBuffer[BUF_SIZE];
    PCONNECTION pIncomingConnection;
    PCONNECTION pOutgoingConnection;
    // Sign of closing an incoming connection.
    BOOL        fIncomingConnectionClosed;
    // Sign of closing an outgoing connection.
    BOOL        fOutgoingConnectionClosed;
    // Counter of using this structure. Used for safe structure deletion.
    LONG        lUseCount;
}COMMON_DATA, *PCOMMON_DATA;

// Connection types.
#define INCOMING    1
#define OUTGOING    2

// Gets the pointer to the COMMON_DATA structure for a connection.
#define DATA(a)     ((PCOMMON_DATA)(a)->pvPerConnectionUserData)

// Returns the connection type.
inline DWORD GetType( PCONNECTION pConnection )
{
    if( pConnection == DATA( pConnection )->pIncomingConnection ){ return INCOMING; }
    if( pConnection == DATA( pConnection )->pOutgoingConnection ){ return OUTGOING; }

    return 0;
}

//##############################################################################################################//
//##############################################################################################################//
//##############################################################################################################//
VOID LogError( PSTR pszFunc, DWORD dwError )
{
    HANDLE hFile = CreateFile( TEXT("ProxyEchoServer.err"), GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, 
						       OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );
    if( hFile == INVALID_HANDLE_VALUE ){ return; }

    SetFilePointer( hFile, 0, NULL, FILE_END );

    CHAR szError[64] = {0};
    wsprintfA( szError, "%s: %d\r\n", pszFunc, dwError );

    DWORD dwWrite = 0;
    WriteFile( hFile, szError, lstrlenA( szError ), &dwWrite, NULL );

    CloseHandle( hFile );
}

//###############################################################################################################//
BOOL CALLBACK OnConnect( PCONNECTION pConnection )
{
    InterlockedIncrement( &g_lOnConnectCount );

    pConnection->pvPerConnectionUserData = malloc( sizeof( COMMON_DATA ) );
    if( NULL == pConnection->pvPerConnectionUserData ){
        LogError( "malloc()", 0 );
        return FALSE;
        }

    // The COMMON_DATA structure is used with this connection.
    DATA( pConnection )->lUseCount = 1;
    // Saving the pointer to the incoming connection.
    DATA( pConnection )->pIncomingConnection = pConnection;
    // Incoming connection established.
    DATA( pConnection )->fIncomingConnectionClosed = FALSE;
    // Outgoing connection not established.
    DATA( pConnection )->fOutgoingConnectionClosed = TRUE;

    // Establishing an asynchronous connection.
    DWORD dwRet = PServer.Connect( g_szServerIP, g_dwServerPort, pConnection->pvPerConnectionUserData, NULL );
    if( dwRet != 0 ){
        LogError( "Connect()", dwRet );
        return FALSE;
        }

    return TRUE;
}

//###############################################################################################################//
BOOL CALLBACK OnOutgoingConnect( PCONNECTION pConnection, DWORD dwNtStatus )
{
    if( 0 /* STATUS_SUCCESS */ == dwNtStatus ){
        InterlockedIncrement( &g_lOnOutgoingConnectCount );

        // The COMMON_DATA structure is used with this connection.
        InterlockedIncrement( &DATA( pConnection )->lUseCount );

        // Saving the pointer to the outgoing connection.
        DATA( pConnection )->pOutgoingConnection = pConnection;

        // Outgoing connection established.
        DATA( pConnection )->fOutgoingConnectionClosed = FALSE;

        // Receiving data from client.
        DWORD dwRet = PServer.PostRecv( DATA( pConnection )->pIncomingConnection, 
                                        DATA( pConnection )->bIncomingBuffer,
                                        BUF_SIZE,
                                        NULL );
        if( dwRet != ERROR_SUCCESS ){
            LogError( "PostRecv()", dwRet );
            return FALSE;
            }
        }
    else{
        // Closing an incoming connection.
        PServer.CloseConnection( DATA( pConnection )->pIncomingConnection );
        return FALSE;
        }

    return TRUE;
}

//##############################################################################################################//
VOID CALLBACK OnDisconnect( PCONNECTION pConnection )
{
    switch( GetType( pConnection ) ){
        case INCOMING:
            InterlockedExchange( (PLONG) &DATA( pConnection )->fIncomingConnectionClosed, TRUE );

            // Closing an outgoing connection.
            if( InterlockedExchange( (PLONG) &DATA( pConnection )->fOutgoingConnectionClosed, TRUE ) == FALSE ){
                PServer.CloseConnection( DATA( pConnection )->pOutgoingConnection );
                }

            InterlockedIncrement( &g_lOnDisconnectCount );
            break;

        case OUTGOING:
            InterlockedExchange( (PLONG) &DATA( pConnection )->fOutgoingConnectionClosed, TRUE );

            // Closing an incoming connection.
            if( InterlockedExchange( (PLONG) &DATA( pConnection )->fIncomingConnectionClosed, TRUE ) == FALSE ){
                PServer.CloseConnection( DATA( pConnection )->pIncomingConnection );
                }

            InterlockedIncrement( &g_lOnOutgoingDisconnectCount );
            break;
        }
}

//##############################################################################################################//
VOID CALLBACK OnRemoveConnection( PCONNECTION pConnection )
{
    switch( GetType( pConnection ) ){
        case INCOMING:
            InterlockedIncrement( &g_lOnRemoveConnectionCount );
            break;

        case OUTGOING:
            InterlockedIncrement( &g_lOnOutgoingRemoveConnectionCount );
            break;
        }
        
    // If the COMMON_DATA structure is no longer used, we are deleting it.
    if( InterlockedDecrement( &DATA( pConnection )->lUseCount ) == 0 ){
        if( pConnection->pvPerConnectionUserData != NULL ){
            free( pConnection->pvPerConnectionUserData );
            }
        }
}

//##############################################################################################################//
BOOL CALLBACK OnRecv( PCONNECTION pConnection, PIO_DATA pIoData )
{
    PCONNECTION pAnotherConnection = NULL;
    PBYTE       pbCurrentBuffer    = NULL;

    switch( GetType( pConnection ) ){
        case INCOMING:
            pAnotherConnection = DATA( pConnection )->pOutgoingConnection;
            pbCurrentBuffer    = DATA( pConnection )->bIncomingBuffer;
            break;

        case OUTGOING:
            pAnotherConnection = DATA( pConnection )->pIncomingConnection;
            pbCurrentBuffer    = DATA( pConnection )->bOutgoingBuffer;
            break;
        }

    // Transferring data to another connection.
    DWORD dwRet = PServer.PostSend( pAnotherConnection,
                                    pbCurrentBuffer,
                                    pIoData->dwBytesTransferred,
                                    NULL );
    if( dwRet != ERROR_SUCCESS ){
        LogError( "PostSend()", dwRet );
        return FALSE;
        }

    return TRUE;
}

//##############################################################################################################//
BOOL CALLBACK OnSend( PCONNECTION pConnection, PIO_DATA pIoData )
{
    switch( GetType( pConnection ) ){
        case INCOMING:
            // The data has been sent to the client - closing the incoming connection.
            return FALSE;

        case OUTGOING:{
            // Receiving data from server.
            DWORD dwRet = PServer.PostRecv( pConnection, 
                                            DATA( pConnection )->bOutgoingBuffer,
                                            BUF_SIZE,
                                            NULL );
            if( dwRet != ERROR_SUCCESS ){
                LogError( "PostRecv()", dwRet );
                return FALSE;
                }
            break;
            }
        }

    return TRUE;
}

//##############################################################################################################//
VOID CALLBACK OnError( PCWSTR pszFunc, DWORD dwError )
{
    CHAR szFunc[64] = {0};
    WideCharToMultiByte( CP_ACP, 0, pszFunc, -1, szFunc, sizeof( szFunc ), NULL, NULL );
    LogError( szFunc, dwError );
}

//###############################################################################################################//
//###############################################################################################################//
//###############################################################################################################//
void DlgMain_OnCommand( HWND hWnd, int id, HWND hWndCtl, UINT codeNotify )
{
	switch( id ){
        case ID_BUTTON_START:{
            SetDlgItemText( hWnd, ID_STATIC_STATUS, TEXT("Starting...") );

            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_START       ), FALSE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_REMOTE_SERVER ), FALSE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_REMOTE_PORT   ), FALSE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_LOCAL_PORT    ), FALSE );

            GetDlgItemTextA( hWnd, ID_EDIT_REMOTE_SERVER, g_szServerIP, sizeof( g_szServerIP ) / sizeof( g_szServerIP[0] ) );
            g_dwServerPort = GetDlgItemInt( hWnd, ID_EDIT_REMOTE_PORT, NULL, FALSE );

            g_lOnConnectCount                  = 0;
            g_lOnOutgoingConnectCount          = 0;
            g_lOnDisconnectCount               = 0;
            g_lOnOutgoingDisconnectCount       = 0;
            g_lOnRemoveConnectionCount         = 0;
            g_lOnOutgoingRemoveConnectionCount = 0;

            PSERVERPARAM param = {0};

            /*
            // Available in the full version.
            SYSTEM_INFO  sysinfo = {0};
            GetSystemInfo( &sysinfo );
            param.dwWorkerThreadsMin = sysinfo.dwNumberOfProcessors * 2;
            param.dwWorkerThreadsMax = sysinfo.dwNumberOfProcessors * 8;
            */

            param.dwWorkerThreadsIdleTimeout   = 60000;
            param.dwUserThreadsMin             = 0;
            param.dwUserThreadsMax             = 0;
            param.dwUserThreadsIdleTimeout     = 60000;
            param.dwCpuUsageMin                = 70;
            param.dwCpuUsageMax                = 90;
            param.dwMultipleIoReqsMax          = 1;
            param.dwPort                       = GetDlgItemInt( hWnd, ID_EDIT_LOCAL_PORT, NULL, FALSE );;
            param.dwInterface                  = INADDR_ANY;
            param.dwSocketRecvBufferSize       = -1;
            param.dwSocketSendBufferSize       = -1;

            /*
              Since the incoming client connection will be idle in the process
              of connecting to the server (PServer_Connect), we should either
              disable the Idle check mechanism or temporarily disable the check
              for each connection using the PServer_EnableConnectionIdleCheck
              function.
            */
            param.dwConnectionsIdleTimeout     = 0;

            param.dwInitAccepts                = 500;
            param.CALLBACKS.OnConnect          = OnConnect;
            param.CALLBACKS.OnDisconnect       = OnDisconnect;
            param.CALLBACKS.OnRemoveConnection = OnRemoveConnection;
            param.CALLBACKS.OnRecv             = OnRecv;
            param.CALLBACKS.OnSend             = OnSend;
            param.CALLBACKS.OnOutgoingConnect  = OnOutgoingConnect;
            param.CALLBACKS.OnError            = OnError;

            DWORD dwRet = PServer.Start( &param );
            if( dwRet != ERROR_SUCCESS ){
                TCHAR szError[64] = {0};
                wsprintf( szError, TEXT("Start(): %d"), dwRet );
                SetDlgItemText( hWnd, ID_STATIC_STATUS, szError );
                EnableWindow( GetDlgItem( hWnd, ID_BUTTON_START       ), TRUE );
                EnableWindow( GetDlgItem( hWnd, ID_EDIT_REMOTE_SERVER ), TRUE );
                EnableWindow( GetDlgItem( hWnd, ID_EDIT_REMOTE_PORT   ), TRUE );
                EnableWindow( GetDlgItem( hWnd, ID_EDIT_LOCAL_PORT    ), TRUE );
                break;
                }

            SetTimer( hWnd, 1, 1000, NULL );

            SetDlgItemText( hWnd, ID_STATIC_STATUS, TEXT("Started") );

            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_STOP ), TRUE );

            g_fStarted = TRUE;
            break;
            }

        case ID_BUTTON_STOP:
            SetDlgItemText( hWnd, ID_STATIC_STATUS, TEXT("Stopping...") );

            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_STOP ), FALSE );

            PServer.Stop();

            KillTimer( hWnd, 1 );

            SetDlgItemText( hWnd, ID_STATIC_CPUUSAGE,                   NULL );
            SetDlgItemText( hWnd, ID_STATIC_WORKER_THREADS,             NULL );
            SetDlgItemText( hWnd, ID_STATIC_WORKER_THREADS_BUSY,        NULL );
            SetDlgItemText( hWnd, ID_STATIC_WORKER_IOLEN,               NULL );
            SetDlgItemText( hWnd, ID_STATIC_USER_THREADS,               NULL );
            SetDlgItemText( hWnd, ID_STATIC_USER_THREADS_BUSY,          NULL );
            SetDlgItemText( hWnd, ID_STATIC_USER_IOLEN,                 NULL );
            SetDlgItemText( hWnd, ID_STATIC_CONNECTIONS,                NULL );
            SetDlgItemText( hWnd, ID_STATIC_OUTGOING_CONNECTIONS,       NULL );
            SetDlgItemText( hWnd, ID_STATIC_ONCONNECT,                  NULL );
            SetDlgItemText( hWnd, ID_STATIC_ONOUTGOINGCONNECT,          NULL );
            SetDlgItemText( hWnd, ID_STATIC_ONDISCONNECT,               NULL );
            SetDlgItemText( hWnd, ID_STATIC_ONOUTGOINGDISCONNECT,       NULL );
            SetDlgItemText( hWnd, ID_STATIC_ONREMOVECONNECTION,         NULL );
            SetDlgItemText( hWnd, ID_STATIC_ONOUTGOINGREMOVECONNECTION, NULL );

            SetDlgItemText( hWnd, ID_STATIC_STATUS, TEXT("Stopped") );

            EnableWindow( GetDlgItem( hWnd, ID_BUTTON_START       ), TRUE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_REMOTE_SERVER ), TRUE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_REMOTE_PORT   ), TRUE );
            EnableWindow( GetDlgItem( hWnd, ID_EDIT_LOCAL_PORT    ), TRUE );
   
            g_fStarted = FALSE;
            break;

        case ID_CHECK_ONTOP:
            SetWindowPos( hWnd, IsDlgButtonChecked( hWnd, ID_CHECK_ONTOP ) ? HWND_TOPMOST : HWND_NOTOPMOST,
                          0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE );
            break;
        }
}

//###############################################################################################################//
BOOL DlgMain_OnInitDialog( HWND hWnd, HWND hWndFocus, LPARAM lParam )
{
    g_hwndMain = hWnd;

    SetDlgItemText( hWnd, ID_EDIT_REMOTE_SERVER, TEXT("127.0.0.1") );
    SetDlgItemInt( hWnd, ID_EDIT_REMOTE_PORT, 6060, FALSE );
    SetDlgItemInt( hWnd, ID_EDIT_LOCAL_PORT, 7070, FALSE );

    return TRUE;
}

//###############################################################################################################//
void DlgMain_OnTimer( HWND hWnd, UINT id )
{
    INFORMATION info = {0};

    PServer.QueryInformation( &info );

    SetDlgItemInt( hWnd, ID_STATIC_CPUUSAGE,                   info.dwCpuUsage,                    FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_WORKER_THREADS,             info.dwWorkerThreadsCount,          FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_WORKER_THREADS_BUSY,        info.dwWorkerThreadsBusyCount,      FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_WORKER_IOLEN,               info.dwIoCompletionLen,             FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_USER_THREADS,               info.dwUserThreadsCount,            FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_USER_THREADS_BUSY,          info.dwUserThreadsBusyCount,        FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_USER_IOLEN,                 info.dwUserIoCompletionLen,         FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_CONNECTIONS,                info.dwConnectionsCount,            FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_OUTGOING_CONNECTIONS,       info.dwOutgoingConnectionsCount,    FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_ONCONNECT,                  g_lOnConnectCount,                  FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_ONOUTGOINGCONNECT,          g_lOnOutgoingConnectCount,          FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_ONDISCONNECT,               g_lOnDisconnectCount,               FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_ONOUTGOINGDISCONNECT,       g_lOnOutgoingDisconnectCount,       FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_ONREMOVECONNECTION,         g_lOnRemoveConnectionCount,         FALSE );
    SetDlgItemInt( hWnd, ID_STATIC_ONOUTGOINGREMOVECONNECTION, g_lOnOutgoingRemoveConnectionCount, FALSE );
}

//###############################################################################################################//
void DlgMain_OnClose( HWND hWnd )
{
    if( g_fStarted ){
        SendMessage( hWnd, WM_COMMAND, ID_BUTTON_STOP, NULL );
        }

	EndDialog( hWnd, 0 );
}

//###############################################################################################################//
int WINAPI Dialog_Main( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
    switch( message ){
        Dlg_Proc( hWnd, WM_INITDIALOG, DlgMain_OnInitDialog );
        Dlg_Proc( hWnd, WM_COMMAND,    DlgMain_OnCommand    );
        Dlg_Proc( hWnd, WM_TIMER,      DlgMain_OnTimer      );
        Dlg_Proc( hWnd, WM_CLOSE,      DlgMain_OnClose      );
        }

    return 0;
}

//##############################################################################################################//
int WINAPI _tWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow )
{
    DialogBox( hInstance, MAKEINTRESOURCE( IDD_DIALOG_MAIN ), NULL, Dialog_Main );

    return 0;
}

//##############################################################################################################//